# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from collections import OrderedDict


NUM_FMT = "{:.4f}"


def _table_format(data, headers=None, index=None, extra_spaces=0, h_bars=None):
    if headers is not None:
        data.insert(0, headers)
    if index is not None:
        index.insert(0, "")
        for idx, row in zip(index, data):
            row.insert(0, idx)

    column_widths = np.asarray([[len(str(v)) for v in row] for row in data]).max(axis=0)

    row_fmt = (
        " | ".join(["{:>%d}" % (w + extra_spaces) for w in column_widths][1:]) + "\n"
    )
    if index is not None:
        row_fmt = "{:<%d} | " % (column_widths[0] + extra_spaces) + row_fmt

    output = ""
    for i, row in enumerate(data):
        if h_bars is not None and i in h_bars:
            output += row_fmt.format(
                *["-" * (w + extra_spaces) for w in column_widths]
            ).replace("|", "+")
        output += row_fmt.format(*row)
    return output


class Result:
    """
    Result Class for a single model

    Parameters
    ----------
    model_name: string, required
        The name of the recommender model.

    metric_avg_results: :obj:`OrderedDict`, required
        A dictionary containing the average result per-metric.

    metric_user_results: :obj:`OrderedDict`, required
        A dictionary containing the average result per-user across different metrics.
    """

    def __init__(self, model_name, metric_avg_results, metric_user_results):
        self.model_name = model_name
        self.metric_avg_results = metric_avg_results
        self.metric_user_results = metric_user_results

    def __str__(self):
        headers = list(self.metric_avg_results.keys())
        data = [[NUM_FMT.format(v) for v in self.metric_avg_results.values()]]
        return _table_format(data, headers, index=[self.model_name], h_bars=[1])


class CVResult(list):
    """
    Cross Validation Result Class for a single model. A list of :obj:`cornac.experiment.Result`.

    Parameters
    ----------
    model_name: string, required
        The name of the recommender model.
        
    Attributes
    ----------    
    metric_mean: :obj:`OrderedDict`
        A dictionary containing the mean of results across all folds per-metric.

    metric_std: :obj:`OrderedDict`
        A dictionary containing the standard deviation of results across all folds per-metric.
    """

    def __init__(self, model_name):
        super().__init__()
        self.model_name = model_name
        self.metric_mean = OrderedDict()
        self.metric_std = OrderedDict()

    def __str__(self):
        return "[{}]\n{}".format(self.model_name, self.table)

    def organize(self):
        headers = list(self[0].metric_avg_results.keys())
        data, index = [], []
        for f, r in enumerate(self):
            data.append([r.metric_avg_results[m] for m in headers])
            index.append("Fold %d" % f)

        data = np.asarray(data)
        mean, std = data.mean(axis=0), data.std(axis=0)
        
        for m, mean_val, std_val in zip(headers, mean, std):
            self.metric_mean[m] = mean_val
            self.metric_std[m] = std_val
        
        data = np.vstack([data, mean, std])
        data = [[NUM_FMT.format(v) for v in row] for row in data]
        index.extend(["Mean", "Std"])
        self.table = _table_format(data, headers, index, h_bars=[1, len(data) - 1])

class PSTResult(list):
    """
    Propensity Stratified Result Class for a single model

    Parameters
    ----------
    model_name: string, required
        The name of the recommender model.
    """

    def __init__(self, model_name):
        super().__init__()
        self.model_name = model_name

    def __str__(self):
        return "[{}]\n{}".format(self.model_name, self.table)

    def organize(self):

        headers = list(self[0].metric_avg_results.keys())

        data, index, sizes = [], [], []
        for f, r in enumerate(self):
            data.append([r.metric_avg_results[m] for m in headers])
            if f == 0:
                index.append("Closed")
            elif f == 1:
                index.append("IPS")
            else:
                index.append("Q%d" % (f - 1))
            sizes.append(r.metric_avg_results["SIZE"])

        # add mean and std rows (total accumulative)
        data = np.asarray(data)
        mean, std = data.mean(axis=0), data.std(axis=0)

        # add unbiased stratified evaluation
        weights = np.asarray(sizes) / sizes[0]
        unbiased = np.average(data[2:], axis=0, weights=weights[2:]) * sum(weights[2:])

        # weighted average does not meaningful for size
        for idx, header in enumerate(headers):
            if header == "SIZE":
                unbiased[idx] = sizes[0]

        # update the table
        data = np.vstack([data, unbiased])
        data = [[NUM_FMT.format(v) for v in row] for row in data]
        index.extend(["Unbiased"])

        # add unbiased to the list
        self.append(
            Result(
                model_name=self[0].model_name,
                metric_avg_results=OrderedDict(zip(headers, unbiased)),
                metric_user_results=None,
            )
        )

        self.table = _table_format(data, headers, index, h_bars=[1, 2, 3, len(data)])


class ExperimentResult(list):
    """
    Result Class for an Experiment. A list of :obj:`cornac.experiment.Result`. 
    """

    def __str__(self):
        headers = list(self[0].metric_avg_results.keys())
        data, index = [], []
        for r in self:
            data.append([NUM_FMT.format(r.metric_avg_results[m]) for m in headers])
            index.append(r.model_name)
        return _table_format(data, headers, index, h_bars=[1])


class CVExperimentResult(ExperimentResult):
    """
    Result Class for a cross-validation Experiment.
    """

    def __str__(self):
        return "\n".join([r.__str__() for r in self])



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np

from ..recommender import Recommender
from ...utils.common import sigmoid
from ...utils.common import scale
from ...exception import ScoreException


class VMF(Recommender):
    """Visual Matrix Factorization.

    Parameters
    ----------
    k: int, optional, default: 10
        The dimension of the user and item factors.
        
    d: int, optional, default: 10
       The dimension of the user visual factors.

    n_epochs: int, optional, default: 100
        The number of epochs for SGD.

    learning_rate: float, optional, default: 0.001
        The learning rate for SGD_RMSProp.

    gamma: float, optional, default: 0.9
        The weight for previous/current gradient in RMSProp.

    lambda_u: float, optional, default: 0.001
        The regularization parameter for user factors.

    lambda_v: float, optional, default: 0.001
        The regularization parameter for item factors.

    lambda_p: float, optional, default: 1.0
        The regularization parameter for user visual factors.

    lambda_e: float, optional, default: 10.
        The regularization parameter for the kernel embedding matrix
        
    lambda_u: float, optional, default: 0.001
        The regularization parameter for user factors.

    name: string, optional, default: 'VMF'
        The name of the recommender model.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model is already \
        pre-trained (The parameters of the model U, V, P, E are not None).
        
    visual_features: See "cornac/examples/vmf_example.py" for an example of how to use \
        cornac's visual modality to load and provide the "item visual features" for VMF.

    verbose: boolean, optional, default: False
        When True, some running logs are displayed.

    init_params: dictionary, optional, default: None
        List of initial parameters, e.g., init_params = {'U':U, 'V':V, 'P': P, 'E': E}.

        U: numpy array of shape (n_users,k), user latent factors.
        V: numpy array of shape (n_items,k), item latent factors.
        P: numpy array of shape (n_users,d), user visual latent factors.
        E: numpy array of shape (d,c), embedding kernel matrix.

    seed: int, optional, default: None
        Random seed for parameters initialization.

    References
    ----------
    * Park, Chanyoung, Donghyun Kim, Jinoh Oh, and Hwanjo Yu. "Do Also-Viewed Products Help User Rating Prediction?."\
     In Proceedings of WWW, pp. 1113-1122. 2017.

    """

    def __init__(
        self,
        name="VMF",
        k=10,
        d=10,
        n_epochs=100,
        batch_size=100,
        learning_rate=0.001,
        gamma=0.9,
        lambda_u=0.001,
        lambda_v=0.001,
        lambda_p=1.0,
        lambda_e=10.0,
        trainable=True,
        verbose=False,
        use_gpu=False,
        init_params=None,
        seed=None,
    ):
        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.d = d
        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.gamma = gamma
        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.lambda_p = lambda_p
        self.lambda_e = lambda_e
        self.use_gpu = use_gpu
        self.loss = np.full(n_epochs, 0)
        self.eps = 0.000000001
        self.seed = seed

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.U = self.init_params.get("U", None)  # user factors
        self.V = self.init_params.get("V", None)  # item factors
        self.P = self.init_params.get("P", None)  # user visual factors
        self.E = self.init_params.get("E", None)  # Kernel embedding matrix

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """

        Recommender.fit(self, train_set, val_set)

        if self.trainable:
            # Item visual cnn-features
            item_features = train_set.item_image.features[: self.train_set.num_items]

            if self.verbose:
                print("Learning...")

            from .vmf import vmf

            res = vmf(
                self.train_set,
                item_features,
                k=self.k,
                d=self.d,
                n_epochs=self.n_epochs,
                batch_size=self.batch_size,
                lambda_u=self.lambda_u,
                lambda_v=self.lambda_v,
                lambda_p=self.lambda_p,
                lambda_e=self.lambda_e,
                learning_rate=self.learning_rate,
                gamma=self.gamma,
                init_params={"U": self.U, "V": self.V, "P": self.P, "E": self.E},
                use_gpu=self.use_gpu,
                verbose=self.verbose,
                seed=self.seed,
            )

            self.U = res["U"]
            self.V = res["V"]
            self.P = res["P"]
            self.E = res["E"]
            self.Q = res["Q"]

            if self.verbose:
                print("Learning completed")

        elif self.verbose:
            print("%s is trained already (trainable = False)" % (self.name))

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.V.dot(self.U[user_idx, :]) + self.Q.dot(
                self.P[user_idx, :]
            )
            # known_item_scores = np.asarray(np.zeros(self.V.shape[0]),dtype='float32')
            # fast_dot(self.U[user_id], self.V, known_item_scores)
            # fast_dot(self.P[user_id], self.Q, known_item_scores)
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )
            user_pred = self.V[item_idx, :].dot(self.U[user_idx, :]) + self.Q[
                item_idx, :
            ].dot(self.P[user_idx, :])
            user_pred = sigmoid(user_pred)

            user_pred = scale(
                user_pred,
                self.train_set.min_rating,
                self.train_set.max_rating,
                0.0,
                1.0,
            )

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from tqdm.auto import trange

from ..recommender import Recommender
from ...exception import ScoreException
from ...utils import get_rng
from ...utils.init_utils import xavier_uniform


class CDL(Recommender):
    """Collaborative Deep Learning.

    Parameters
    ----------
    name: string, default: 'CDL'
        The name of the recommender model.

    k: int, optional, default: 50
        The dimension of the latent factors.

    max_iter: int, optional, default: 100
        Maximum number of iterations or the number of epochs for SGD.

    autoencoder_structure: list, default: None
        The number of neurons of encoder/decoder layer for SDAE.
        For example, autoencoder_structure = [200], the SDAE structure will be [vocab_size, 200, k, 200, vocab_size]

    act_fn: str, default: 'relu'
        Name of the activation function used for the auto-encoder.
        Supported functions: ['sigmoid', 'tanh', 'elu', 'relu', 'relu6', 'leaky_relu', 'identity']

    learning_rate: float, optional, default: 0.001
        The learning rate for AdamOptimizer.

    vocab_size: int, default: 8000
        The size of text input for the SDAE.

    lambda_u: float, optional, default: 0.1
        The regularization parameter for users.

    lambda_v: float, optional, default: 10
        The regularization parameter for items.

    lambda_w: float, optional, default: 0.1
        The regularization parameter for SDAE weights.

    lambda_n: float, optional, default: 1000
        The regularization parameter for SDAE output.

    a: float, optional, default: 1
        The confidence of observed ratings.

    b: float, optional, default: 0.01
        The confidence of unseen ratings.

    corruption_rate: float, optional, default: 0.3
        The corruption ratio for input text of the SDAE.

    dropout_rate: float, optional, default: 0.1
        The probability that each element is removed in dropout of SDAE.

    batch_size: int, optional, default: 128
        The batch size for SGD.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already 
        pre-trained (U and V are not None).

    init_params: dictionary, optional, default: None
        List of initial parameters, e.g., init_params = {'U':U, 'V':V}

        U: ndarray, shape (n_users,k)
            The user latent factors, optional initialization via init_params.
        V: ndarray, shape (n_items,k)
            The item latent factors, optional initialization via init_params.

    seed: int, optional, default: None
        Random seed for weight initialization.

    References
    ----------
    * Hao Wang, Naiyan Wang, Dit-Yan Yeung. CDL: Collaborative Deep Learning for Recommender Systems. In : SIGKDD. 2015. p. 1235-1244.
    """

    def __init__(
        self,
        name="CDL",
        k=50,
        autoencoder_structure=None,
        act_fn="relu",
        lambda_u=0.1,
        lambda_v=10,
        lambda_w=0.1,
        lambda_n=1000,
        a=1,
        b=0.01,
        corruption_rate=0.3,
        learning_rate=0.001,
        vocab_size=8000,
        dropout_rate=0.1,
        batch_size=128,
        max_iter=100,
        trainable=True,
        verbose=True,
        init_params=None,
        seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.lambda_w = lambda_w
        self.lambda_n = lambda_n
        self.a = a
        self.b = b
        self.corruption_rate = corruption_rate
        self.dropout_rate = dropout_rate
        self.learning_rate = learning_rate
        self.vocab_size = vocab_size
        self.name = name
        self.max_iter = max_iter
        self.autoencoder_structure = autoencoder_structure
        self.act_fn = act_fn
        self.batch_size = batch_size
        self.verbose = verbose
        self.seed = seed
        self.rng = get_rng(seed)

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.U = self.init_params.get("U", None)
        self.V = self.init_params.get("V", None)

    def _init(self):
        n_users, n_items = self.train_set.num_users, self.train_set.num_items

        if self.U is None:
            self.U = xavier_uniform((n_users, self.k), self.rng)
        if self.V is None:
            self.V = xavier_uniform((n_items, self.k), self.rng)

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self._init()

        if self.trainable:
            self._fit_cdl()

        return self

    def _fit_cdl(self):
        import tensorflow.compat.v1 as tf
        from .cdl import Model
        
        tf.disable_eager_execution()

        R = self.train_set.csc_matrix  # csc for efficient slicing over items
        n_users, n_items = self.train_set.num_users, self.train_set.num_items

        text_feature = self.train_set.item_text.batch_bow(
            np.arange(n_items)
        )  # bag of word feature
        text_feature = (text_feature - text_feature.min()) / (
            text_feature.max() - text_feature.min()
        )  # normalization

        # Build model
        layer_sizes = (
            [self.vocab_size]
            + self.autoencoder_structure
            + [self.k]
            + self.autoencoder_structure
            + [self.vocab_size]
        )
        tf.set_random_seed(self.seed)
        model = Model(
            n_users=n_users,
            n_items=n_items,
            n_vocab=self.vocab_size,
            k=self.k,
            layers=layer_sizes,
            lambda_u=self.lambda_u,
            lambda_v=self.lambda_v,
            lambda_w=self.lambda_w,
            lambda_n=self.lambda_n,
            lr=self.learning_rate,
            dropout_rate=self.dropout_rate,
            U=self.U,
            V=self.V,
            act_fn=self.act_fn,
            seed=self.seed,
        )

        # Training model
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        with tf.Session(config=config) as sess:
            sess.run(tf.global_variables_initializer())

            loop = trange(self.max_iter, disable=not self.verbose)
            for _ in loop:
                corruption_mask = self.rng.binomial(
                    1, 1 - self.corruption_rate, size=(n_items, self.vocab_size)
                )
                sum_loss = 0
                count = 0
                for i, batch_ids in enumerate(
                    self.train_set.item_iter(self.batch_size, shuffle=True)
                ):
                    batch_R = R[:, batch_ids]
                    batch_C = np.ones(batch_R.shape) * self.b
                    batch_C[batch_R.nonzero()] = self.a

                    feed_dict = {
                        model.text_mask: corruption_mask[batch_ids, :],
                        model.text_input: text_feature[batch_ids],
                        model.ratings: batch_R.A,
                        model.C: batch_C,
                        model.item_ids: batch_ids,
                    }
                    sess.run(model.opt1, feed_dict)  # train U, V
                    _, _loss = sess.run(
                        [model.opt2, model.loss], feed_dict
                    )  # train SDAE

                    sum_loss += _loss
                    count += len(batch_ids)
                    if i % 10 == 0:
                        loop.set_postfix(loss=(sum_loss / count))

            self.U, self.V = sess.run([model.U, model.V])

        tf.reset_default_graph()

        if self.verbose:
            print("Learning completed!")

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items
        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.V.dot(self.U[user_idx, :])
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )
            user_pred = self.V[item_idx, :].dot(self.U[user_idx, :])
            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import multiprocessing

import numpy as np
from scipy.sparse import csr_matrix, coo_matrix

from ..recommender import Recommender
from ...exception import ScoreException
from ...utils import get_rng
from ...utils.fast_sparse_funcs import inplace_csr_row_normalize_l2
from .similarity import compute_similarity, compute_score, compute_score_single


EPS = 1e-8

SIMILARITIES = ["cosine", "pearson"]
WEIGHTING_OPTIONS = ["idf", "bm25"]


def _mean_centered(ui_mat):
    """Subtract every rating values with mean value of the corresponding rows"""
    mean_arr = np.zeros(ui_mat.shape[0])
    for i in range(ui_mat.shape[0]):
        start_idx, end_idx = ui_mat.indptr[i : i + 2]
        mean_arr[i] = np.mean(ui_mat.data[start_idx:end_idx])
        row_data = ui_mat.data[start_idx:end_idx]
        row_data -= mean_arr[i]
        row_data[row_data == 0] = EPS
        ui_mat.data[start_idx:end_idx] = row_data

    return ui_mat, mean_arr


def _amplify(ui_mat, alpha=1.0):
    """Exponentially amplify values of similarity matrix"""
    if alpha == 1.0:
        return ui_mat

    for i, w in enumerate(ui_mat.data):
        ui_mat.data[i] = w ** alpha if w > 0 else -(-w) ** alpha
    return ui_mat


def _idf_weight(ui_mat):
    """Weight the matrix Inverse Document (Item) Frequency"""
    X = coo_matrix(ui_mat)

    # calculate IDF
    N = float(X.shape[0])
    idf = np.log(N / np.bincount(X.col))

    weights = idf[ui_mat.indices] + EPS
    return weights


def _bm25_weight(ui_mat):
    """Weight the matrix with BM25 algorithm"""
    K1 = 1.2
    B = 0.8

    X = coo_matrix(ui_mat)
    X.data = np.ones_like(X.data)

    N = float(X.shape[0])
    idf = np.log(N / np.bincount(X.col))

    # calculate length_norm per document (user)
    row_sums = np.ravel(X.sum(axis=1))
    average_length = row_sums.mean()
    length_norm = (1.0 - B) + B * row_sums / average_length

    # bm25 weights
    weights = (K1 + 1.0) / (K1 * length_norm[X.row] + X.data) * idf[X.col] + EPS
    return weights


class UserKNN(Recommender):
    """User-Based Nearest Neighbor.

    Parameters
    ----------
    name: string, default: 'UserKNN'
        The name of the recommender model.

    k: int, optional, default: 20
        The number of nearest neighbors.
       
    similarity: str, optional, default: 'cosine'
        The similarity measurement. Supported types: ['cosine', 'pearson']
    
    mean_centered: bool, optional, default: False
        Whether values of the user-item rating matrix will be centered by the mean
        of their corresponding rows (mean rating of each user).  
    
    weighting: str, optional, default: None
        The option for re-weighting the rating matrix. Supported types: ['idf', 'bm25'].
        If None, no weighting is applied.
          
    amplify: float, optional, default: 1.0
        Amplifying the influence on similarity weights.
        
    num_threads: int, optional, default: 0
        Number of parallel threads for training. If num_threads=0, all CPU cores will be utilized.
        If seed is not None, num_threads=1 to remove randomness from parallelization.

    seed: int, optional, default: None
        Random seed for weight initialization.

    References
    ----------
    * CarlKadie, J. B. D. (1998). Empirical analysis of predictive algorithms for collaborative filtering. Microsoft Research Microsoft Corporation One Microsoft Way Redmond, WA, 98052.
    * Aggarwal, C. C. (2016). Recommender systems (Vol. 1). Cham: Springer International Publishing.
    """

    def __init__(
        self,
        name="UserKNN",
        k=20,
        similarity="cosine",
        mean_centered=False,
        weighting=None,
        amplify=1.0,
        num_threads=0,
        trainable=True,
        verbose=True,
        seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.similarity = similarity
        self.mean_centered = mean_centered
        self.weighting = weighting
        self.amplify = amplify
        self.seed = seed
        self.rng = get_rng(seed)

        if self.similarity not in SIMILARITIES:
            raise ValueError(
                "Invalid similarity choice, supported {}".format(SIMILARITIES)
            )

        if self.weighting is not None and self.weighting not in WEIGHTING_OPTIONS:
            raise ValueError(
                "Invalid weighting choice, supported {}".format(WEIGHTING_OPTIONS)
            )

        if seed is not None:
            self.num_threads = 1
        elif num_threads > 0 and num_threads < multiprocessing.cpu_count():
            self.num_threads = num_threads
        else:
            self.num_threads = multiprocessing.cpu_count()

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self.ui_mat = self.train_set.matrix.copy()
        self.mean_arr = np.zeros(self.ui_mat.shape[0])
        if self.train_set.min_rating != self.train_set.max_rating:  # explicit feedback
            self.ui_mat, self.mean_arr = _mean_centered(self.ui_mat)

        if self.mean_centered or self.similarity == "pearson":
            weight_mat = self.ui_mat.copy()
        else:
            weight_mat = self.train_set.matrix.copy()

        # re-weighting
        if self.weighting == "idf":
            weight_mat.data *= np.sqrt(_idf_weight(self.train_set.matrix))
        elif self.weighting == "bm25":
            weight_mat.data *= np.sqrt(_bm25_weight(self.train_set.matrix))

        # only need item-user matrix for prediction
        self.iu_mat = self.ui_mat.T.tocsr()
        del self.ui_mat

        self.sim_mat = compute_similarity(
            weight_mat, k=self.k, num_threads=self.num_threads, verbose=self.verbose
        )
        self.sim_mat = _amplify(self.sim_mat, self.amplify)

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items
        """
        if self.train_set.is_unk_user(user_idx):
            raise ScoreException(
                "Can't make score prediction for (user_id=%d)" % user_idx
            )

        if item_idx is not None and self.train_set.is_unk_item(item_idx):
            raise ScoreException(
                "Can't make score prediction for (item_id=%d)" % item_idx
            )

        if item_idx is not None:
            weighted_avg = compute_score_single(
                True,
                self.sim_mat[user_idx].A.ravel(),
                self.iu_mat.indptr[item_idx],
                self.iu_mat.indptr[item_idx + 1],
                self.iu_mat.indices,
                self.iu_mat.data,
                k=self.k,
            )
            return self.mean_arr[user_idx] + weighted_avg

        weighted_avg = np.zeros(self.train_set.num_items)
        compute_score(
            True,
            self.sim_mat[user_idx].A.ravel(),
            self.iu_mat.indptr,
            self.iu_mat.indices,
            self.iu_mat.data,
            k=self.k,
            num_threads=self.num_threads,
            output=weighted_avg,
        )
        known_item_scores = self.mean_arr[user_idx] + weighted_avg

        return known_item_scores


class ItemKNN(Recommender):
    """Item-Based Nearest Neighbor.

    Parameters
    ----------
    name: string, default: 'ItemKNN'
        The name of the recommender model.

    k: int, optional, default: 20
        The number of nearest neighbors.

    similarity: str, optional, default: 'cosine'
        The similarity measurement. Supported types: ['cosine', 'pearson']
      
    mean_centered: bool, optional, default: False
        Whether values of the user-item rating matrix will be centered by the mean
        of their corresponding rows (mean rating of each user).  
         
    weighting: str, optional, default: None
        The option for re-weighting the rating matrix. Supported types: ['idf', 'bm25'].
        If None, no weighting is applied.
               
    amplify: float, optional, default: 1.0
        Amplifying the influence on similarity weights.
         
    num_threads: int, optional, default: 0
        Number of parallel threads for training. If num_threads=0, all CPU cores will be utilized.
        If seed is not None, num_threads=1 to remove randomness from parallelization.

    seed: int, optional, default: None
        Random seed for weight initialization.

    References
    ----------
    * Sarwar, B., Karypis, G., Konstan, J., & Riedl, J. (2001, April). Item-based collaborative filtering recommendation algorithms. In Proceedings of the 10th international conference on World Wide Web (pp. 285-295).
    * Aggarwal, C. C. (2016). Recommender systems (Vol. 1). Cham: Springer International Publishing.
    """

    def __init__(
        self,
        name="ItemKNN",
        k=20,
        similarity="cosine",
        mean_centered=False,
        weighting=None,
        amplify=1.0,
        num_threads=0,
        trainable=True,
        verbose=True,
        seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.similarity = similarity
        self.mean_centered = mean_centered
        self.weighting = weighting
        self.amplify = amplify
        self.seed = seed
        self.rng = get_rng(seed)

        if self.similarity not in SIMILARITIES:
            raise ValueError(
                "Invalid similarity choice, supported {}".format(SIMILARITIES)
            )

        if self.weighting is not None and self.weighting not in WEIGHTING_OPTIONS:
            raise ValueError(
                "Invalid weighting choice, supported {}".format(WEIGHTING_OPTIONS)
            )

        if seed is not None:
            self.num_threads = 1
        elif num_threads > 0 and num_threads < multiprocessing.cpu_count():
            self.num_threads = num_threads
        else:
            self.num_threads = multiprocessing.cpu_count()

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self.ui_mat = self.train_set.matrix.copy()
        self.mean_arr = np.zeros(self.ui_mat.shape[0])
        if self.train_set.min_rating != self.train_set.max_rating:  # explicit feedback
            self.ui_mat, self.mean_arr = _mean_centered(self.ui_mat)

        if self.mean_centered:
            weight_mat = self.ui_mat.copy()
        else:
            weight_mat = self.train_set.matrix.copy()

        if self.similarity == "pearson":  # centered by columns
            weight_mat, _ = _mean_centered(weight_mat.T.tocsr())
            weight_mat = weight_mat.T.tocsr()

        # re-weighting
        if self.weighting == "idf":
            weight_mat.data *= np.sqrt(_idf_weight(self.train_set.matrix))
        elif self.weighting == "bm25":
            weight_mat.data *= np.sqrt(_bm25_weight(self.train_set.matrix))

        weight_mat = weight_mat.T.tocsr()
        self.sim_mat = compute_similarity(
            weight_mat, k=self.k, num_threads=self.num_threads, verbose=self.verbose
        )
        self.sim_mat = _amplify(self.sim_mat, self.amplify)

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items
        """
        if self.train_set.is_unk_user(user_idx):
            raise ScoreException(
                "Can't make score prediction for (user_id=%d)" % user_idx
            )

        if item_idx is not None and self.train_set.is_unk_item(item_idx):
            raise ScoreException(
                "Can't make score prediction for (item_id=%d)" % item_idx
            )

        if item_idx is not None:
            weighted_avg = compute_score_single(
                False,
                self.ui_mat[user_idx].A.ravel(),
                self.sim_mat.indptr[item_idx],
                self.sim_mat.indptr[item_idx + 1],
                self.sim_mat.indices,
                self.sim_mat.data,
                k=self.k,
            )
            return self.mean_arr[user_idx] + weighted_avg

        weighted_avg = np.zeros(self.train_set.num_items)
        compute_score(
            False,
            self.ui_mat[user_idx].A.ravel(),
            self.sim_mat.indptr,
            self.sim_mat.indices,
            self.sim_mat.data,
            k=self.k,
            num_threads=self.num_threads,
            output=weighted_avg,
        )
        return self.mean_arr[user_idx] + weighted_avg



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np

from ..recommender import Recommender
from ...utils.common import sigmoid
from ...utils.common import scale
from ...exception import ScoreException


class MCF(Recommender):
    """Matrix Co-Factorization.

    Parameters
    ----------
    k: int, optional, default: 5
        The dimension of the latent factors.

    max_iter: int, optional, default: 100
        Maximum number of iterations or the number of epochs for SGD.

    learning_rate: float, optional, default: 0.001
        The learning rate for SGD_RMSProp.

    gamma: float, optional, default: 0.9
        The weight for previous/current gradient in RMSProp.

    lamda: float, optional, default: 0.001
        The regularization parameter.

    name: string, optional, default: 'MCF'
        The name of the recommender model.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model is already \
        pre-trained (U and V are not None).

    item-affinity network: See "cornac/examples/mcf_office.py" for an example of how to use \
        cornac's graph modality to load and provide the "item-affinity network" for MCF.

    verbose: boolean, optional, default: False
        When True, some running logs are displayed.

    init_params: dictionary, optional, default: None
        List of initial parameters, e.g., init_params = {'U': U, 'V': V, 'Z', Z}.

        U: ndarray, shape (n_users, k)
            User latent factors.
        
        V: ndarray, shape (n_items, k)
            Item latent factors.
        
        Z: ndarray, shape (n_items, k)
            The "Also-Viewed" item latent factors.

    seed: int, optional, default: None
        Random seed for parameters initialization.

    References
    ----------
    * Park, Chanyoung, Donghyun Kim, Jinoh Oh, and Hwanjo Yu. "Do Also-Viewed Products Help User Rating Prediction?."\
     In Proceedings of WWW, pp. 1113-1122. 2017.

    """

    def __init__(
        self,
        k=5,
        max_iter=100,
        learning_rate=0.001,
        gamma=0.9,
        lamda=0.001,
        name="MCF",
        trainable=True,
        verbose=False,
        init_params=None,
        seed=None,
    ):
        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.max_iter = max_iter
        self.learning_rate = learning_rate
        self.gamma = gamma
        self.lamda = lamda
        self.seed = seed

        self.ll = np.full(max_iter, 0)
        self.eps = 0.000000001

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.U = self.init_params.get("U", None)  # matrix of user factors
        self.V = self.init_params.get("V", None)  # matrix of item factors
        self.Z = self.init_params.get("Z", None)  # matrix of Also-Viewed item factors

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        if self.trainable:
            # user-item interactions
            (rat_uid, rat_iid, rat_val) = train_set.uir_tuple

            # item-item affinity network
            map_iid = train_set.item_indices
            (net_iid, net_jid, net_val) = train_set.item_graph.get_train_triplet(
                map_iid, map_iid
            )
            if [self.train_set.min_rating, self.train_set.max_rating] != [0, 1]:
                if self.train_set.min_rating == self.train_set.max_rating:
                    rat_val = scale(rat_val, 0.0, 1.0, 0.0, self.train_set.max_rating)
                else:
                    rat_val = scale(
                        rat_val,
                        0.0,
                        1.0,
                        self.train_set.min_rating,
                        self.train_set.max_rating,
                    )

            if [min(net_val), max(net_val)] != [0, 1]:
                if min(net_val) == max(net_val):
                    net_val = scale(net_val, 0.0, 1.0, 0.0, max(net_val))
                else:
                    net_val = scale(net_val, 0.0, 1.0, min(net_val), max(net_val))

            rat_val = np.array(rat_val, dtype="float32")
            rat_uid = np.array(rat_uid, dtype="int32")
            rat_iid = np.array(rat_iid, dtype="int32")

            net_val = np.array(net_val, dtype="float32")
            net_iid = np.array(net_iid, dtype="int32")
            net_jid = np.array(net_jid, dtype="int32")

            if self.verbose:
                print("Learning...")

            from cornac.models.mcf import mcf

            res = mcf.mcf(
                rat_uid,
                rat_iid,
                rat_val,
                net_iid,
                net_jid,
                net_val,
                k=self.k,
                n_users=train_set.num_users,
                n_items=train_set.num_items,
                n_ratings=len(rat_val),
                n_edges=len(net_val),
                n_epochs=self.max_iter,
                lamda=self.lamda,
                learning_rate=self.learning_rate,
                gamma=self.gamma,
                init_params={"U": self.U, "V": self.V, "Z": self.Z},
                verbose=self.verbose,
                seed=self.seed,
            )

            self.U = np.asarray(res["U"])
            self.V = np.asarray(res["V"])
            self.Z = np.asarray(res["Z"])

            if self.verbose:
                print("Learning completed")
        elif self.verbose:
            print("%s is trained already (trainable = False)" % self.name)

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.V.dot(self.U[user_idx, :])
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            user_pred = self.V[item_idx, :].dot(self.U[user_idx, :])

            user_pred = sigmoid(user_pred)
            if self.train_set.min_rating == self.train_set.max_rating:
                user_pred = scale(user_pred, 0.0, self.train_set.max_rating, 0.0, 1.0)
            else:
                user_pred = scale(
                    user_pred,
                    self.train_set.min_rating,
                    self.train_set.max_rating,
                    0.0,
                    1.0,
                )

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from . import Modality
from collections import OrderedDict


class SentimentModality(Modality):
    """Aspect module
    Parameters
    ----------
    data: List[tuple], required
        A triplet list of user, item, and sentiment information \
        which also a triplet list of aspect, opinion, and sentiment, \
        e.g., data=[('user1', 'item1', [('aspect1', 'opinion1', 'sentiment1')])].
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.raw_data = kwargs.get('data', OrderedDict())

    @property
    def sentiment(self):
        return self.__sentiment

    @sentiment.setter
    def sentiment(self, input_sentiment):
        self.__sentiment = input_sentiment

    @property
    def num_aspects(self):
        """Return the number of aspects"""
        return len(self.aspect_id_map)

    @property
    def num_opinions(self):
        """Return the number of aspects"""
        return len(self.opinion_id_map)

    @property
    def user_sentiment(self):
        return self.__user_sentiment

    @user_sentiment.setter
    def user_sentiment(self, input_user_sentiment):
        self.__user_sentiment = input_user_sentiment

    @property
    def item_sentiment(self):
        return self.__item_sentiment

    @item_sentiment.setter
    def item_sentiment(self, input_item_sentiment):
        self.__item_sentiment = input_item_sentiment

    @property
    def aspect_id_map(self):
        return self.__aspect_id_map

    @aspect_id_map.setter
    def aspect_id_map(self, input_aspect_id_map):
        self.__aspect_id_map = input_aspect_id_map

    @property
    def opinion_id_map(self):
        return self.__opinion_id_map

    @opinion_id_map.setter
    def opinion_id_map(self, input_opinion_id_map):
        self.__opinion_id_map = input_opinion_id_map

    def _build_sentiment(self, uid_map, iid_map, dok_matrix):
        self.user_sentiment = OrderedDict()
        self.item_sentiment = OrderedDict()
        aid_map = OrderedDict()
        oid_map = OrderedDict()
        sentiment = OrderedDict()
        for idx, (raw_uid, raw_iid, sentiment_tuples) in enumerate(self.raw_data):
            user_idx = uid_map.get(raw_uid, None)
            item_idx = iid_map.get(raw_iid, None)
            if user_idx is None or item_idx is None or dok_matrix[user_idx, item_idx] == 0:
                continue
            user_dict = self.user_sentiment.setdefault(user_idx, OrderedDict())
            user_dict[item_idx] = idx
            item_dict = self.item_sentiment.setdefault(item_idx, OrderedDict())
            item_dict[user_idx] = idx

            mapped_tup = []
            for tup in sentiment_tuples:
                aspect, opinion, polarity = tup[0], tup[1], float(tup[2])
                aspect_idx = aid_map.setdefault(aspect, len(aid_map))
                opinion_idx = oid_map.setdefault(opinion, len(oid_map))
                mapped_tup.append((aspect_idx, opinion_idx, polarity))
            sentiment.setdefault(idx, mapped_tup)

        self.sentiment = sentiment
        self.aspect_id_map = aid_map
        self.opinion_id_map = oid_map

    def build(self, uid_map=None, iid_map=None, dok_matrix=None, **kwargs):
        """Build the model based on provided list of ordered ids
        """
        if uid_map is not None and iid_map is not None and dok_matrix is not None:
            self._build_sentiment(uid_map, iid_map, dok_matrix)
        return self



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import scipy.sparse as sp

from cornac.models.c2pf import c2pf
from ..recommender import Recommender


# Recommender class for Collaborative Context Poisson Factorization (C2PF)
class C2PF(Recommender):
    """Collaborative Context Poisson Factorization.

    Parameters
    ----------
    k: int, optional, default: 100
        The dimension of the latent factors.

    max_iter: int, optional, default: 100
        Maximum number of iterations for variational C2PF.

    variant: string, optional, default: 'c2pf'
        C2pf's variant: c2pf: 'c2pf', 'tc2pf' (tied-c2pf) or 'rc2pf' (reduced-c2pf). \
        Please refer to the original paper for details.

    name: string, optional, default: None
        The name of the recommender model. If None, \
        then "variant" is used as the default name of the model.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already \
        pre-trained (Theta, Beta and Xi are not None).

    Item_context: See "cornac/examples/c2pf_example.py" in the GitHub repo for an example of how to use \
        cornac's graph modality to load and provide "item context" for C2PF.

    init_params: dict, optional, default: None
        List of initial parameters, e.g., init_params = {'G_s':G_s, 'G_r':G_r, 'L_s':L_s, 'L_r':L_r, 'L2_s':L2_s, 'L2_r':L2_r, 'L3_s':L3_s, 'L3_r': L3_r}

        Theta: ndarray, shape (n_users, k)
            The expected user latent factors.

        Beta: ndarray, shape (n_items, k)
            The expected item latent factors.

        Xi: ndarray, shape (n_items, k)
            The expected context item latent factors multiplied by context effects Kappa.
            
        G_s: ndarray, shape (n_users, k)
            Represent the "shape" parameters of Gamma distribution over Theta.
            
        G_r: ndarray, shape (n_users, k)
            Represent the "rate" parameters of Gamma distribution over Theta. 
        
        L_s: ndarray, shape (n_items, k)
            Represent the "shape" parameters of Gamma distribution over Beta.
            
        L_r: ndarray, shape (n_items, k)
            Represent the "rate" parameters of Gamma distribution over Beta. 
        
        L2_s: ndarray, shape (n_items, k)
            Represent the "shape" parameters of Gamma distribution over Xi.
            
        L2_r: ndarray, shape (n_items, k)
            Represent the "rate" parameters of Gamma distribution over Xi.
             
        L3_s: ndarray
            Represent the "shape" parameters of Gamma distribution over Kappa.
            
        L3_r: ndarray 
            Represent the "rate" parameters of Gamma distribution over Kappa.

    References
    ----------
    * Salah, Aghiles, and Hady W. Lauw. A Bayesian Latent Variable Model of User Preferences with Item Context. \
    In IJCAI, pp. 2667-2674. 2018.
    """

    def __init__(
        self,
        k=100,
        max_iter=100,
        variant="c2pf",
        name=None,
        trainable=True,
        verbose=False,
        init_params=None,
    ):
        if name is None:
            Recommender.__init__(
                self, name=variant.upper(), trainable=trainable, verbose=verbose
            )
        else:
            Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)

        self.k = k
        self.max_iter = max_iter

        self.ll = np.full(max_iter, 0)
        self.eps = 0.000000001
        # self.aux_info = aux_info  # item-context matrix in the triplet sparse format: (row_id, col_id, value)
        self.variant = variant

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.Theta = self.init_params.get("Theta", None)
        self.Beta = self.init_params.get("Beta", None)
        self.Xi = self.init_params.get("Xi", None)
        self.Gs = self.init_params.get("G_s", None)
        self.Gr = self.init_params.get("G_r", None)
        self.Ls = self.init_params.get("L_s", None)
        self.Lr = self.init_params.get("L_r", None)
        self.L2s = self.init_params.get("L2_s", None)
        self.L2r = self.init_params.get("L2_r", None)
        self.L3s = self.init_params.get("L3_s", None)
        self.L3r = self.init_params.get("L3_r", None)

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        X = sp.csc_matrix(self.train_set.matrix)

        # recover the striplet sparse format from csc sparse matrix X (needed to feed c++)
        (rid, cid, val) = sp.find(X)
        val = np.array(val, dtype="float32")
        rid = np.array(rid, dtype="int32")
        cid = np.array(cid, dtype="int32")
        tX = np.concatenate(
            (np.concatenate(([rid], [cid]), axis=0).T, val.reshape((len(val), 1))),
            axis=1,
        )
        del rid, cid, val

        if self.trainable:
            # use pre-trained params if exists, otherwise from constructor
            init_params = {
                "G_s": self.Gs,
                "G_r": self.Gr,
                "L_s": self.Ls,
                "L_r": self.Lr,
                "L2_s": self.L2s,
                "L2_r": self.L2r,
                "L3_s": self.L3s,
                "L3_r": self.L3r,
            }

            map_iid = train_set.item_indices
            (rid, cid, val) = train_set.item_graph.get_train_triplet(map_iid, map_iid)
            context_info = np.hstack(
                (rid.reshape(-1, 1), cid.reshape(-1, 1), val.reshape(-1, 1))
            )

            if self.variant == "c2pf":
                res = c2pf.c2pf(
                    tX,
                    X.shape[0],
                    X.shape[1],
                    context_info,
                    X.shape[1],
                    X.shape[1],
                    self.k,
                    self.max_iter,
                    init_params,
                )
            elif self.variant == "tc2pf":
                res = c2pf.t_c2pf(
                    tX,
                    X.shape[0],
                    X.shape[1],
                    context_info,
                    X.shape[1],
                    X.shape[1],
                    self.k,
                    self.max_iter,
                    init_params,
                )
            elif self.variant == "rc2pf":
                res = c2pf.r_c2pf(
                    tX,
                    X.shape[0],
                    X.shape[1],
                    context_info,
                    X.shape[1],
                    X.shape[1],
                    self.k,
                    self.max_iter,
                    init_params,
                )
            else:
                res = c2pf.c2pf(
                    tX,
                    X.shape[0],
                    X.shape[1],
                    context_info,
                    X.shape[1],
                    X.shape[1],
                    self.k,
                    self.max_iter,
                    init_params,
                )

            self.Theta = sp.csc_matrix(res["Z"]).todense()
            self.Beta = sp.csc_matrix(res["W"]).todense()
            self.Xi = sp.csc_matrix(res["Q"]).todense()

            # overwrite init_params for future fine-tuning
            self.Gs = np.asarray(res["G_s"])
            self.Gr = np.asarray(res["G_r"])
            self.Ls = np.asarray(res["L_s"])
            self.Lr = np.asarray(res["L_r"])
            self.L2s = np.asarray(res["L2_s"])
            self.L2r = np.asarray(res["L2_r"])
            self.L3s = np.asarray(res["L3_s"])
            self.L3r = np.asarray(res["L3_r"])

        elif self.verbose:
            print("%s is trained already (trainable = False)" % (self.name))

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if self.variant == "c2pf" or self.variant == "tc2pf":
            if item_idx is None:
                user_pred = (
                    self.Beta * self.Theta[user_idx, :].T
                    + self.Xi * self.Theta[user_idx, :].T
                )
            else:
                user_pred = (
                    self.Beta[item_idx, :] * self.Theta[user_idx, :].T
                    + self.Xi * self.Theta[user_idx, :].T
                )
        elif self.variant == "rc2pf":
            if item_idx is None:
                user_pred = self.Xi * self.Theta[user_idx, :].T
            else:
                user_pred = self.Xi[item_idx,] * self.Theta[user_idx, :].T
        else:
            if item_idx is None:
                user_pred = (
                    self.Beta * self.Theta[user_idx, :].T
                    + self.Xi * self.Theta[user_idx, :].T
                )
            else:
                user_pred = (
                    self.Beta[item_idx, :] * self.Theta[user_idx, :].T
                    + self.Xi * self.Theta[user_idx, :].T
                )
        # transform user_pred to a flatten array,
        user_pred = np.array(user_pred, dtype="float64").flatten()

        return user_pred
